(************** Content-type: application/mathematica **************

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     17418,        495]*)
(*NotebookOutlinePosition[     18079,        518]*)
(*  CellTagsIndexPosition[     18035,        514]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["\<\

Examples of the use of the NonCommutative Control Systems package\
\>", \
"Title"],

Cell[CellGroupData[{

Cell["\<\

(* :Title: \tNC-CSP-Demo.nb *)


<<NCAlgebra.m
<<ControlSystems`
<<NCControl.m



Clear[\.19a1, a2,a3,a4,b1,b2,c1,c2,d,u1,u2,u3,u4,v1,v2,w1,w2,z, 
\t  B, F, G, H, X, Y, Z, W,  T, U, S];

SetNonCommutative[a1,a2,a3,a4,b1,b2,c1,c2,d,u1,u2,u3,u4,v1,v2,w1,w2,z ];

SetCommutative[B, F, G, H, X, Y, Z, W, S];

sys1 = StateSpace[ {{a1,a2},{a3,a4}},{{b1},{b2}},{{c1,c2}},{{d}} ];
sys2 = StateSpace[ {{u1,u2},{u3,u4}},{{v1},{v2}},{{w1,w2}},{{z}} ];
csys1 = StateSpace[{{B, F}, {G, H}}, {{X}, { Y}}, {{Z, W}},  {{S}}]; \
\>", \
"Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False],

Cell[BoxData[
    \("You have already loaded NCAlgebra.m"\)], "Print"],

Cell[BoxData[
    \(" ******************************************************** "\)], "Print"],

Cell[BoxData[
    \(" * NOTE: In order to ensure that TransferFunction[]     * "\)], "Print"],

Cell[BoxData[
    \(" *       works for systems with potentially non-        * "\)], "Print"],

Cell[BoxData[
    \(" *       commutative arguments, the option              * "\)], "Print"],

Cell[BoxData[
    \(" *             ReductionMethod->NCInverse               * "\)], "Print"],

Cell[BoxData[
    \(" *       is now the TransferFunction default.           * "\)], "Print"],

Cell[BoxData[
    \(" *       As TransferFunction[] may generate non-        * "\)], "Print"],

Cell[BoxData[
    \(" *       standard (though correct) output for systems   * "\)], "Print"],

Cell[BoxData[
    \(" *       with commutative arguments, you may wish, in   * "\)], "Print"],

Cell[BoxData[
    \(" *       such cases, to replace the call                * "\)], "Print"],

Cell[BoxData[
    \(" *              TransferFunction[X]                     * "\)], "Print"],

Cell[BoxData[
    \(" *       with                                           * "\)], "Print"],

Cell[BoxData[
    \(" *       TransferFunction[X,                            * "\)], "Print"],

Cell[BoxData[
    \(" *              ReductionMethod->DeterminantExpansion], * "\)], "Print"],

Cell[BoxData[
    \(" *       which is the CSP II default.                   * "\)], "Print"],

Cell[BoxData[
    \(" ******************************************************** "\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["SYSTEM INTERCONNECTIONS", "Subtitle"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(\[IndentingNewLine]\)\(\[IndentingNewLine]\)\(SeriesConnect[sys1, 
        sys2] // TraditionalForm\)\)\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {"a1", "a2", "0", "0", "b1"},
                    {"a3", "a4", "0", "0", "b2"},
                    {\(v1 ** c1\), \(v1 ** c2\), "u1", "u2", \(v1 ** d\)},
                    {\(v2 ** c1\), \(v2 ** c2\), "u3", "u4", \(v2 ** d\)},
                    {\(z ** c1\), \(z ** c2\), "w1", "w2", \(z ** d\)}
                    },
                  RowLines->{False, False, False, True, False},
                  ColumnLines->{False, False, False, True, False},
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox["\[Bullet]",
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalS]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (StateSpace[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(FeedbackConnect[sys1, sys2] // 
      TraditionalForm\)\(\ \)\(\[IndentingNewLine]\)
    \)\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {\(a1 - b1 ** z ** \(inv(d ** z + 1)\) ** c1\), \(a2 - 
                        b1 ** z ** \(inv(d ** z + 1)\) ** 
                          c2\), \(-b1 ** \((1 - 
                              z ** \(inv(d ** z + 1)\) ** d)\) ** 
                          w1\), \(-b1 ** \((1 - 
                              z ** \(inv(d ** z + 1)\) ** d)\) ** 
                          w2\), \(b1 - 
                        b1 ** z ** \(inv(d ** z + 1)\) ** d\)},
                    {\(a3 - b2 ** z ** \(inv(d ** z + 1)\) ** c1\), \(a4 - 
                        b2 ** z ** \(inv(d ** z + 1)\) ** 
                          c2\), \(-b2 ** \((1 - 
                              z ** \(inv(d ** z + 1)\) ** d)\) ** 
                          w1\), \(-b2 ** \((1 - 
                              z ** \(inv(d ** z + 1)\) ** d)\) ** 
                          w2\), \(b2 - 
                        b2 ** z ** \(inv(d ** z + 1)\) ** d\)},
                    {\(v1 ** \(inv(d ** z + 1)\) ** 
                        c1\), \(v1 ** \(inv(d ** z + 1)\) ** c2\), \(u1 - 
                        v1 ** \(inv(d ** z + 1)\) ** d ** w1\), \(u2 - 
                        v1 ** \(inv(d ** z + 1)\) ** d ** w2\), \(v1 ** \(inv(
                          d ** z + 1)\) ** d\)},
                    {\(v2 ** \(inv(d ** z + 1)\) ** 
                        c1\), \(v2 ** \(inv(d ** z + 1)\) ** c2\), \(u3 - 
                        v2 ** \(inv(d ** z + 1)\) ** d ** w1\), \(u4 - 
                        v2 ** \(inv(d ** z + 1)\) ** d ** w2\), \(v2 ** \(inv(
                          d ** z + 1)\) ** d\)},
                    {\(c1 - d ** z ** \(inv(d ** z + 1)\) ** c1\), \(c2 - 
                        d ** z ** \(inv(d ** z + 1)\) ** 
                          c2\), \(-d ** \((1 - 
                              z ** \(inv(d ** z + 1)\) ** d)\) ** 
                          w1\), \(-d ** \((1 - 
                              z ** \(inv(d ** z + 1)\) ** d)\) ** w2\), \(d - 
                        d ** z ** \(inv(d ** z + 1)\) ** d\)}
                    },
                  RowLines->{False, False, False, True, False},
                  ColumnLines->{False, False, False, True, False},
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox["\[Bullet]",
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalS]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (StateSpace[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(ParallelConnect[sys1, sys2] // 
      TraditionalForm\)\(\ \)\(\[IndentingNewLine]\)
    \)\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {"a1", "a2", "0", "0", "b1"},
                    {"a3", "a4", "0", "0", "b2"},
                    {"0", "0", "u1", "u2", "v1"},
                    {"0", "0", "u3", "u4", "v2"},
                    {"c1", "c2", "w1", "w2", \(d + z\)}
                    },
                  RowLines->{False, False, False, True, False},
                  ColumnLines->{False, False, False, True, False},
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox["\[Bullet]",
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalS]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (StateSpace[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["NEW SYSTEMS FROM OLD", "Subtitle"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(NCInverseSystem[sys1] // 
      TraditionalForm\)\(\ \)\(\[IndentingNewLine]\)
    \)\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {\(a1 - b1 ** \(inv(d)\) ** c1\), \(a2 - 
                        b1 ** \(inv(d)\) ** c2\), \(-b1 ** \(inv(d)\)\)},
                    {\(a3 - b2 ** \(inv(d)\) ** c1\), \(a4 - 
                        b2 ** \(inv(d)\) ** c2\), \(-b2 ** \(inv(d)\)\)},
                    {\(\(inv(d)\) ** c1\), \(\(inv(d)\) ** c2\), \(inv(d)\)}
                    },
                  RowLines->{False, True, False},
                  ColumnLines->{False, True, False},
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox["\[Bullet]",
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalS]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (StateSpace[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(NCConjugateSystem[sys1] // TraditionalForm\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {\(-\(tp(a1)\)\), \(-\(tp(a3)\)\), \(-\(tp(c1)\)\)},
                    {\(-\(tp(a2)\)\), \(-\(tp(a4)\)\), \(-\(tp(c2)\)\)},
                    {\(tp(b1)\), \(tp(b2)\), \(tp(d)\)}
                    },
                  RowLines->{False, True, False},
                  ColumnLines->{False, True, False},
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox["\[Bullet]",
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalS]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (StateSpace[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["TRANSFER FUNCTION", "Subtitle"],

Cell["Non-Commutative System", "Subsubtitle"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TransferFunction[sys1]\  // \ TraditionalForm\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {\(d + \((c2 ** \(inv(\(-a4\) + \[ScriptS] - 
                                    a3 ** \(inv(\[ScriptS] - a1)\) ** a2)\) + 
                              c1 ** \(inv(\[ScriptS] - a1)\) ** 
                                a2 ** \(inv(\(-a4\) + \[ScriptS] - 
                                    a3 ** \(inv(\[ScriptS] - a1)\) ** 
                                      a2)\))\) ** 
                          b2 + \((c1 ** \((inv(\[ScriptS] - 
                                        a1) + \(inv(\[ScriptS] - a1)\) ** 
                                      a2 ** \(inv(\(-a4\) + \[ScriptS] - 
                                        a3 ** \(inv(\[ScriptS] - a1)\) ** 
                                        a2)\) ** 
                                      a3 ** \(inv(\[ScriptS] - a1)\))\) + 
                              c2 ** \(inv(\(-a4\) + \[ScriptS] - 
                                    a3 ** \(inv(\[ScriptS] - a1)\) ** a2)\) ** 
                                a3 ** \(inv(\[ScriptS] - a1)\))\) ** b1\)}
                    },
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[InvisibleSpace]\"\>",
                ShowStringCharacters->False],
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalT]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (TransferFunction[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Commutative System", "Subsubtitle"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TransferFunction[csys1]\  // \ TraditionalForm\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {\(S + 
                        Y\ \((W\/\(\(-\(\(F\ G\)\/\(\[ScriptS] - B\)\)\) - H \
+ \[ScriptS]\) + \(F\ Z\)\/\(\((\[ScriptS] - B)\)\ \((\(-\(\(F\ G\)\/\(\
\[ScriptS] - B\)\)\) - H + \[ScriptS])\)\))\) + 
                        X\ \((\(G\ W\)\/\(\((\[ScriptS] - B)\)\ \((\(-\(\(F\ \
G\)\/\(\[ScriptS] - B\)\)\) - H + \[ScriptS])\)\) + 
                              Z\ \((\(F\ G\)\/\(\((\[ScriptS] - B)\)\^2\ \
\((\(-\(\(F\ G\)\/\(\[ScriptS] - B\)\)\) - H + \[ScriptS])\)\) + 
                                    1\/\(\[ScriptS] - B\))\))\)\)}
                    },
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[InvisibleSpace]\"\>",
                ShowStringCharacters->False],
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalT]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (TransferFunction[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TransferFunction[csys1, \ 
        ReductionMethod \[Rule] \ DeterminantExpansion]\  // \ 
      TraditionalForm\)], "Input"],

Cell[BoxData[
    FormBox[
      TagBox[
        FormBox[
          SubsuperscriptBox[
            FormBox[
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {\(S + 
                        Y\ \((W\/\(\(-\(\(F\ G\)\/\(\[ScriptS] - B\)\)\) - H \
+ \[ScriptS]\) + \(F\ Z\)\/\(\((\[ScriptS] - B)\)\ \((\(-\(\(F\ G\)\/\(\
\[ScriptS] - B\)\)\) - H + \[ScriptS])\)\))\) + 
                        X\ \((\(G\ W\)\/\(\((\[ScriptS] - B)\)\ \((\(-\(\(F\ \
G\)\/\(\[ScriptS] - B\)\)\) - H + \[ScriptS])\)\) + 
                              Z\ \((\(F\ G\)\/\(\((\[ScriptS] - B)\)\^2\ \
\((\(-\(\(F\ G\)\/\(\[ScriptS] - B\)\)\) - H + \[ScriptS])\)\) + 
                                    1\/\(\[ScriptS] - B\))\))\)\)}
                    },
                  AllowScriptLevelChange->False], "\[NoBreak]", ")"}],
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[InvisibleSpace]\"\>",
                ShowStringCharacters->False],
              "TraditionalForm"], 
            FormBox[
              StyleBox["\<\"\[ScriptCapitalT]\"\>",
                Editable->False,
                ShowStringCharacters->False],
              "TraditionalForm"],
            MultilineFunction->None],
          "TraditionalForm"],
        (TransferFunction[ 
          SlotSequence[ 1]]&)], TraditionalForm]], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.1 for X",
ScreenRectangle->{{0, 1280}, {0, 1024}},
AutoGeneratedPackage->Automatic,
WindowSize->{784, 700},
WindowMargins->{{Automatic, 8}, {Automatic, 0}}
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1727, 52, 93, 4, 300, "Title"],

Cell[CellGroupData[{
Cell[1845, 60, 621, 25, 312, "Input",
  InitializationCell->True],
Cell[2469, 87, 70, 1, 23, "Print"],
Cell[2542, 90, 93, 1, 23, "Print"],
Cell[2638, 93, 93, 1, 23, "Print"],
Cell[2734, 96, 93, 1, 23, "Print"],
Cell[2830, 99, 93, 1, 23, "Print"],
Cell[2926, 102, 93, 1, 23, "Print"],
Cell[3022, 105, 93, 1, 23, "Print"],
Cell[3118, 108, 93, 1, 23, "Print"],
Cell[3214, 111, 93, 1, 23, "Print"],
Cell[3310, 114, 93, 1, 23, "Print"],
Cell[3406, 117, 93, 1, 23, "Print"],
Cell[3502, 120, 93, 1, 23, "Print"],
Cell[3598, 123, 93, 1, 23, "Print"],
Cell[3694, 126, 93, 1, 23, "Print"],
Cell[3790, 129, 93, 1, 23, "Print"],
Cell[3886, 132, 93, 1, 23, "Print"],
Cell[3982, 135, 93, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4112, 141, 43, 0, 64, "Subtitle"],

Cell[CellGroupData[{
Cell[4180, 145, 140, 2, 59, "Input"],
Cell[4323, 149, 1144, 27, 113, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5504, 181, 128, 3, 43, "Input"],
Cell[5635, 186, 2858, 55, 113, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8530, 246, 128, 3, 43, "Input"],
Cell[8661, 251, 1079, 27, 113, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[9789, 284, 40, 0, 64, "Subtitle"],

Cell[CellGroupData[{
Cell[9854, 288, 122, 3, 43, "Input"],
Cell[9979, 293, 1144, 27, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[11160, 325, 75, 1, 27, "Input"],
Cell[11238, 328, 997, 25, 77, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[12284, 359, 37, 0, 64, "Subtitle"],
Cell[12324, 361, 45, 0, 57, "Subsubtitle"],

Cell[CellGroupData[{
Cell[12394, 365, 78, 1, 27, "Input"],
Cell[12475, 368, 1815, 37, 43, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14327, 410, 41, 0, 57, "Subsubtitle"],

Cell[CellGroupData[{
Cell[14393, 414, 79, 1, 27, "Input"],
Cell[14475, 417, 1353, 31, 77, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15865, 453, 145, 3, 27, "Input"],
Cell[16013, 458, 1353, 31, 77, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

